"""Module for working with locations."""


def cluster_locations(locations: dict) -> dict:
    """Cluster locations which appears on same location.

    Args:
        locations (dict): Dictionary mapping IDs to location class.

    Returns:
        dict: Dictionary mapping (x, y) coord tuple to list of locations.
    """
    coord_to_locations = {}

    for key, loc in locations.items():
        if not loc.get_2d() in coord_to_locations:
            coord_to_locations[loc.get_2d()] = []
        coord_to_locations[loc.get_2d()].append(key)

    return coord_to_locations



"""Parser of Excel data files."""
from openpyxl import load_workbook
from xlrd import open_workbook

from virtual_warehouse.parser.data_model import (
    Inventory,
    Item,
    ItemUnit,
    Location,
    Order,
)
from virtual_warehouse.parser.utils import estimate_sheet_type


class Document:
    """Document class which loads xls or xlsx file and parse different data objects."""

    def __init__(self, file_path):
        # Determines backend for loading documents (xlsx files uses openpyxl)
        self.is_xlsx = Document.check_xlsx(file_path)
        self.locations = {}
        self.items = {}
        self.balance = {}
        self.orders = {}
        if self.is_xlsx:
            # self.doc = load_workbook(data_path, read_only=True, keep_links=False)
            self.doc = open_workbook(file_path)
        else:
            self.doc = open_workbook(file_path)

    @staticmethod
    def check_xlsx(file_path):
        """Check if document is .xlsx document (required for openpyxl library)."""
        return file_path[-5:] == ".xlsx"

    @staticmethod
    def get_sheet_names(file_path):
        """Get names of all sheets in document."""
        if Document.check_xlsx(file_path):
            doc = load_workbook(file_path, read_only=True, keep_links=False)
            names = doc.sheetnames
        else:
            doc = open_workbook(file_path, on_demand=True)
            names = doc.sheet_names()
        return [[n, estimate_sheet_type(n)] for n in names]

    def parse_locations(self, sheet_name="LOCATIONmaster"):
        """Parse LOCATIONmaster sheet."""
        sheet = self.doc.sheet_by_name(sheet_name)
        for row in range(1, sheet.nrows):
            location_id = str(sheet.cell(row, 0).value)
            if not location_id:
                continue

            self.locations[location_id] = Location.create(
                *(sheet.cell(row, i).value for i in range(11))
            )

        return self.locations

    def parse_coordinates(self, sheet_name="XYZ_coordinates"):
        """Parse XYZ_coordinates sheet."""
        sheet = self.doc.sheet_by_name(sheet_name)
        for row in range(1, sheet.nrows):
            location_id = str(sheet.cell(row, 0).value)
            if not location_id:
                continue

            self.locations[location_id].set_coord(
                *(sheet.cell(row, i).value for i in range(1, 4))
            )

        return self.locations

    def parse_items(self, sheet_name="ITEMmaster"):
        """Parse ITEMmaster sheet."""
        sheet = self.doc.sheet_by_name(sheet_name)
        for row in range(1, sheet.nrows):
            item_id, description, gtype, zone = (
                sheet.cell(row, i).value for i in range(4)
            )
            item_id = str(item_id)
            if not item_id:
                continue

            unit_levels = []
            for col in range(4, sheet.ncols, 8):
                unit_levels.append(
                    ItemUnit.create(
                        f"{item_id}-u{col}",
                        *(sheet.cell(row, col + i).value for i in range(8)),
                    )
                )
            self.items[item_id] = Item.create(
                item_id, description, gtype, zone, unit_levels[0], unit_levels
            )

        return self.items

    def parse_inventory_balance(self, sheet_name="Inventory Ballance"):
        """Parse Inventory Balance sheet  ('balance' in final version, most likely)."""
        sheet = self.doc.sheet_by_name(sheet_name)
        for row in range(1, sheet.nrows):
            date, location_id = sheet.cell(row, 0).value, sheet.cell(row, 1).value
            location_id = str(location_id)
            if not date:
                continue

            if date not in self.balance:
                self.balance[date] = {}
            self.balance[date][location_id] = Inventory.create(
                *(sheet.cell(row, i).value for i in range(10))
            )

        return self.balance

    def parse_orders(self, sheet_name="Order"):
        """Parse Order sheet."""
        sheet = self.doc.sheet_by_name(sheet_name)
        for row in range(1, sheet.nrows):
            order_id = str(sheet.cell(row, 0).value)
            if not order_id:
                continue

            if order_id in self.orders:
                self.orders[order_id].add_item(
                    *(sheet.cell(row, i).value for i in range(7, 11))
                )
            else:
                self.orders[order_id] = Order.create(
                    *(sheet.cell(row, i).value for i in range(11))
                )

        return self.orders

    def parse_document(self):
        """Parse the whole document."""
        locations = self.parse_locations()
        locations = self.parse_coordinates()
        items = self.parse_items()
        balance = self.parse_inventory_balance()
        orders = self.parse_orders()

        return locations, items, balance, orders



"""Module providing ViewController class - connector class between QML and Python."""
from PySide2.QtCore import Property, QObject, Qt, QThread, QUrl, Signal, Slot

from virtual_warehouse.heatmap import calculate_frquencies
from virtual_warehouse.location_models import (
    MultiLocation,
    SingleLocation,
    UniversalLocationListModel,
)
from virtual_warehouse.location_utils import cluster_locations
from virtual_warehouse.map import Map
from virtual_warehouse.parser.data_model import Inventory, Item, Location, Order
from virtual_warehouse.parser.excel_parser import Document
from virtual_warehouse.tab_controller import (
    HoverListModel,
    TabItem,
    TabLocation,
    TabOrder,
    UniversalListModel,
)


class DataLoaderThread(QThread):
    """Thread which loads warehouse data from file."""

    locationsReady = Signal(object)
    itemsReady = Signal(object)
    inventoryReady = Signal(object)
    ordersReady = Signal(object)
    frequenciesReady = Signal()

    def __init__(
        self,
        file_path,
        sheet_types,
        locations=None,
        items=None,
        inventory=None,
        orders=None,
    ):
        """Initialize thread params for loading file in separate thread.

        Args:
            file_path (str): path to file for loading
            sheet_types (list[dict[str, str]]): list of dictionaries, each containing
                keys "name": name of the sheet, "type": name of the type.
                There are 6 types ("None", "Locations", "Coordinates", "Items",
                "Inventory", "Orders")
            locations (dict[str, Location]): previously loaded locations
            items (dict[str, Item]): previously loaded items
            inventory (dict[str, Inventory]): previously loaded inventory
            orders (dict[str, Inventory]): previously loaded orders
        """
        super(DataLoaderThread, self).__init__()
        self.file_path = file_path
        self.sheets = sheet_types

        self.locations = locations
        self.items = items
        self.inventory = inventory
        self.orders = orders

    def run(self):
        """Load data from file path and emit data through signal."""

        def where(arr, y):
            return [x["name"] for x in arr if x["type"] == y]

        document = Document(QUrl(self.file_path).toLocalFile())

        if (
            len(where(self.sheets, "Locations")) > 0
            and len(where(self.sheets, "Coordinates")) > 0
        ):
            Location.destroy_all()
            for sheet in where(self.sheets, "Locations"):
                self.locations = document.parse_locations(sheet)

            for sheet in where(self.sheets, "Coordinates"):
                self.locations = document.parse_coordinates(sheet)

            self.locationsReady.emit(self.locations)

        if len(where(self.sheets, "Items")) > 0:
            Item.destroy_all()
            for sheet in where(self.sheets, "Items"):
                self.items = document.parse_items(sheet)

            self.itemsReady.emit(self.items)

        if len(where(self.sheets, "Inventory")) > 0:
            Inventory.destroy_all()
            for sheet in where(self.sheets, "Inventory"):
                self.inventory = document.parse_inventory_balance(sheet)

            self.inventoryReady.emit(self.inventory)

        if len(where(self.sheets, "Orders")) > 0:
            Order.destroy_all()
            for sheet in where(self.sheets, "Orders"):
                self.orders = document.parse_orders(sheet)

            self.ordersReady.emit(self.orders)

        # First query is always slower than rest (probably some initialization going on)
        Item.get_by_locations([self.locations[next(iter(self.locations))]])

        calculate_frquencies(self.locations, self.inventory, self.orders)
        self.frequenciesReady.emit()


class QueryThread(QThread):
    """Thread runs database queries."""

    dataReady = Signal(object)

    def __init__(self, function, *args):
        """Initialize thread params for running the query.

        Args:
            function (Callable): function which runs the query.
            *args: arguments which will be passed to the function
        """
        super(QueryThread, self).__init__()
        self.function = function
        self.args = args

    def run(self):
        """Run the query and emit data through dataReady signal."""
        data = self.function(*self.args)
        self.dataReady.emit(data)


class ViewController(QObject):
    """Main controller which communicates with QML GUI."""

    def __init__(self, parent=None):
        super(ViewController, self).__init__(parent)

        self._is2D = True
        self._is_heatmap = False

        self._model3D = UniversalLocationListModel(on_change=self.modelChanged.emit)
        self._model2D = UniversalLocationListModel(on_change=self.modelChanged.emit)
        self._map = Map()

        self._sidebar_model = HoverListModel(TabLocation)
        self._location_model = UniversalListModel(TabLocation)
        self._item_model = UniversalListModel(TabItem)
        self._order_model = UniversalListModel(TabOrder)

        self.selected_idxs = {}

        self.locations = None
        self.items = None
        self.inventory = None
        self.orders = None

        self._query_thread = None
        self._loader = None
        self._progress_value = 1

    modelChanged = Signal()
    sidebarChanged = Signal()
    drawModeChanged = Signal()
    itemSelected = Signal()
    progressChanged = Signal()

    @Property(QObject, constant=False, notify=modelChanged)
    def map(self):
        return self._map

    @Property(QObject, constant=False, notify=modelChanged)
    def model(self):
        return self._model2D if self._is2D else self._model3D

    @Property(QObject, constant=False, notify=modelChanged)
    def model2D(self):
        return self._model2D

    @Property(QObject, constant=False, notify=modelChanged)
    def model3D(self):
        return self._model3D

    @Property(QObject, constant=False, notify=modelChanged)
    def location_model(self):
        return self._location_model

    @Property(QObject, constant=False, notify=sidebarChanged)
    def sidebar_model(self):
        return self._sidebar_model

    @Property(QObject, constant=False, notify=modelChanged)
    def item_model(self):
        return self._item_model

    @Property(QObject, constant=False, notify=modelChanged)
    def order_model(self):
        return self._order_model

    @Property(bool, constant=False, notify=drawModeChanged)
    def is_heatmap(self):
        return self._is_heatmap

    @Property(float, constant=False, notify=progressChanged)
    def progress_value(self):
        return self._progress_value

    @progress_value.setter
    def set_progress_value(self, val):
        self._progress_value = val
        self.progressChanged.emit()

    @Slot(result=bool)
    def is2D(self):
        return self._is2D

    @Slot()
    def switch_heatmap(self):
        self._is_heatmap = not self._is_heatmap
        self.drawModeChanged.emit()

    @Slot()
    def switch_view(self):
        """Switch 2D - 3D model and update selection."""
        self._is2D = not self._is2D
        self.selected_idxs.clear()
        for name in self._location_model._checked:
            idx = self.model.name_to_idx[name]
            self.selected_idxs[idx] = self.selected_idxs.get(idx, 0) + 1
        self.itemSelected.emit()

    def _select_locations(self, selected, checked, clear):
        """Update selected locations in the map.

        Args:
            selected (list[str]): list of names of selected locations
            checked (bool): whether we should add or remove the locations from selection
            clear (bool): if true previously selected indexes will be cleared
        """
        if clear:
            self.selected_idxs.clear()

        for name in selected:
            idx = self.model.name_to_idx[name]
            # Add or remove one occurrence of locations from counter
            self.selected_idxs[idx] = self.selected_idxs.get(idx, 0) + (
                1 if checked else -1
            )
            if self.selected_idxs[idx] <= 0:
                self.selected_idxs.pop(idx)

        self.itemSelected.emit()

    @Slot(str, bool)
    def select_location(self, selected, checked):
        """Select location index of location checked/unchecked in tab list.

        Args:
            selected (str): name of selected location
            checked (bool): check state, if true add location, remove otherwise
        """
        self._select_locations([selected], checked, clear=False)

    @Slot(int, bool)
    def select_map_location(self, idx, control=False):
        """Update selected location from map (CTRL adds location)."""
        if not control:
            self.selected_idxs.clear()

        if idx >= 0:
            names = self.model.get_idx(idx).names
            self.selected_idxs[idx] = len(names)
            self._location_model.set_checked(names, control)
            self._sidebar_model.set_selected(names)
        else:
            self._location_model.set_checked([])
            self._sidebar_model.set_selected([])
        self.itemSelected.emit()

    @Slot(bool)
    def select_all(self, select):
        """Select/unselect all locations from the tab list in the map."""
        self._select_locations(self.location_model._selected, select, clear=True)

    # Connecting sidebar tabs
    def _connect_tabs(self, src_tab_model, dst_tab_model, connector, locations=False):
        """Run query and update tab model with resulting data."""

        def callback(data):
            """Process data from query (callback function)."""
            names = [i[0].name for i in data]
            dst_tab_model.set_checked(names)
            self._query_thread = None
            if locations:
                self._select_locations(names, checked=True, clear=True)

        if src_tab_model._checked:
            objs = [src_tab_model._objects[k]._i for k in src_tab_model._checked]
            if self._query_thread is None:
                self._query_thread = QueryThread(connector, objs)
                self._query_thread.dataReady.connect(callback, Qt.QueuedConnection)
                self._query_thread.start()
        else:
            dst_tab_model.set_checked([])

    @Slot()
    def checked_locations_to_items(self):
        """Connect locations -> items tabs."""
        self._connect_tabs(
            self._location_model, self._item_model, Item.get_by_locations
        )

    @Slot()
    def checked_orders_to_items(self):
        """Connect orders -> items tabs."""
        self._connect_tabs(self._order_model, self._item_model, Item.get_by_orders)

    @Slot()
    def checked_items_to_orders(self):
        """Connect items -> orders tabs."""
        self._connect_tabs(self._item_model, self._order_model, Order.get_by_items)

    @Slot()
    def checked_locations_to_orders(self):
        """Connect locations -> orders tabs."""
        self._connect_tabs(
            self._location_model, self._order_model, Order.get_by_locations
        )

    @Slot()
    def checked_orders_to_locations(self):
        """Connect orders -> locations tabs."""
        self._connect_tabs(
            self._order_model, self._location_model, Location.get_by_orders, True
        )

    @Slot()
    def checked_items_to_locations(self):
        """Connect items -> locations tabs."""
        self._connect_tabs(
            self._item_model, self._location_model, Location.get_by_items, True
        )

    @Slot(int)
    def hover_item(self, idx):
        """Hover location on the map - displaying statistics in sidebar."""
        if idx >= 0:
            names = self.model.get_idx(idx).names
            self._sidebar_model.set_hovered(names, True)
        else:
            self._sidebar_model.set_hovered([], False)
        self.sidebarChanged.emit()

    @Slot(result="QVariantList")
    def get_selected(self):
        """Get list of indexes of selected locations inside the map."""
        return list(self.selected_idxs.keys())

    @Slot(str, result="QVariantList")
    def get_sheets(self, file_path):
        """Get names of sheets inside the file and expected type.

        Args:
            file_path (str): file url

        Returns:
            list[tuple[str, str]]: list containing lists [[name, type_name], ...]
        """
        return Document.get_sheet_names(QUrl(file_path).toLocalFile())

    @Slot(str, "QVariantList")
    def load(self, file_path, types):
        """Load excel file with corresponding sheet types.

        Args:
            file_path (str): file url should be processed with QUrl(file_path)
            types (list[dict[str, str]]): list of dictionaries, each containing
                keys "name": name of the sheet, "type": name of the type.
                There are 6 types ("None", "Locations", "Coordinates", "Items",
                "Inventory", "Orders")
        """
        self.progress_value = 0
        self._loader = DataLoaderThread(
            file_path, types, self.locations, self.items, self.inventory, self.orders
        )
        self._loader.locationsReady.connect(self._load_locations, Qt.QueuedConnection)
        self._loader.itemsReady.connect(self._load_items, Qt.QueuedConnection)
        self._loader.inventoryReady.connect(self._load_inventory, Qt.QueuedConnection)
        self._loader.ordersReady.connect(self._load_orders, Qt.QueuedConnection)
        self._loader.frequenciesReady.connect(
            self._load_frequencies, Qt.QueuedConnection
        )
        self._loader.start()

    def _load_locations(self, locations):
        """Process loaded locations (callback function)."""
        self.selected_idxs.clear()
        self.locations = locations

        self._map.set_data(locations)

        self._location_model.clear_checked()
        self._location_model.set_data(locations)
        self._location_model.set_selected(
            [k for k, v in locations.items() if v.has_ltype == "rack"]
        )

        self._sidebar_model.set_data(locations)

        self._model3D.set_data(
            {k: SingleLocation(v) for k, v in self.locations.items()}
        )

        clusters = cluster_locations(self.locations)
        multi_loc = {}
        for k, v in clusters.items():
            multi_loc[k] = MultiLocation([self.locations[l] for l in v])
        self._model2D.set_data(multi_loc)

        self.modelChanged.emit()
        self.progress_value = 0.3

    def _load_items(self, items):
        """Process loaded items (callback function)."""
        self.items = items
        self._item_model.clear_checked()
        self._item_model.set_data(items)
        self._item_model.set_selected(list(items.keys()))
        self.progress_value = 0.5

    def _load_inventory(self, inventory):
        """Process loaded inventory (callback function)."""
        self.inventory = inventory
        self.progress_value = 0.7

    def _load_orders(self, orders):
        """Process loaded orders (callback function)."""
        self.orders = orders
        self._order_model.clear_checked()
        self._order_model.set_data(orders)
        self._order_model.set_selected(list(orders.keys()))
        self.progress_value = 0.9

    def _load_frequencies(self):
        """Update frequencies (callback function)."""
        self._model2D.update_max_heat()
        self._model3D.update_max_heat()
        self.progress_value = 1
        self.drawModeChanged.emit()



"""Module providing basic functions for working with warehouse map."""
from PySide2.QtCore import Property, QObject


class Map(QObject):
    """Object holding basic informations about the map."""

    def __init__(self, locations=None):
        QObject.__init__(self)
        if locations:
            self.set_data(locations)
        else:
            self._min_x = 0
            self._max_x = 0
            self._min_y = 0
            self._max_y = 0
            self._min_z = 0
            self._max_z = 0

    def set_data(self, locations):
        self._min_x = min(l.has_x for l in locations.values())
        self._max_x = max(l.has_x + l.has_width for l in locations.values())

        self._min_y = min(l.has_y for l in locations.values())
        self._max_y = max(l.has_y + l.has_length for l in locations.values())

        self._min_z = min(l.has_z for l in locations.values())
        self._max_z = max(l.has_z + l.has_height for l in locations.values())

    @Property(float, constant=True)
    def min_x(self):
        return self._min_x

    @Property(float, constant=True)
    def max_x(self):
        return self._max_x

    @Property(float, constant=True)
    def min_y(self):
        return self._min_y

    @Property(float, constant=True)
    def max_y(self):
        return self._max_y

    @Property(float, constant=True)
    def min_z(self):
        return self._min_z

    @Property(float, constant=True)
    def max_z(self):
        return self._max_z



"""Module calculating heat map and statistics on top of warehouse model."""


def item_locations(inventory):
    """Create dictionary mapping (date, item) to list of locations."""
    item_to_loc = {}
    for date, status in inventory.items():
        item_to_loc[date] = {}
        for loc_id, loc_inv in status.items():
            for item in loc_inv.has_items:
                if item.name not in item_to_loc[date]:
                    item_to_loc[date][item.name] = []
                item_to_loc[date][item.name].append(loc_id)

    return item_to_loc


def calculate_frquencies(locations, inventory, orders):
    """Calculate order frequency for each location."""
    item_locs = item_locations(inventory)

    date = list(item_locs.keys())[-1]

    for order in orders.values():
        for item in order.has_ordered_items:
            for loc in item_locs[date][item.has_item.name]:
                locations[loc].has_freq += item.has_total_qty


# Bytes representing values of viridis colors for 0 to 255
VIRIDIS_COLORS = b'D\x01TD\x02UD\x03WE\x05XE\x06ZE\x08[F\t\\F\x0b^F\x0c_F\x0eaG\x0fbG\x11cG\x12eG\x14fG\x15gG\x16iG\x18jH\x19kH\x1alH\x1cnH\x1doH\x1epH qH!rH"sH#tG%uG&vG\'wG(xG*yG+zG,{F-|F/|F0}F1~E2\x7fE4\x7fE5\x80E6\x81D7\x81D9\x82C:\x83C;\x83C<\x84B=\x84B>\x85B@\x85AA\x86AB\x86@C\x87@D\x87?E\x87?G\x88>H\x88>I\x89=J\x89=K\x89=L\x89<M\x8a<N\x8a;P\x8a;Q\x8a:R\x8b:S\x8b9T\x8b9U\x8b8V\x8b8W\x8c7X\x8c7Y\x8c6Z\x8c6[\x8c5\\\x8c5]\x8c4^\x8d4_\x8d3`\x8d3a\x8d2b\x8d2c\x8d1d\x8d1e\x8d1f\x8d0g\x8d0h\x8d/i\x8d/j\x8d.k\x8e.l\x8e.m\x8e-n\x8e-o\x8e,p\x8e,q\x8e,r\x8e+s\x8e+t\x8e*u\x8e*v\x8e*w\x8e)x\x8e)y\x8e(z\x8e(z\x8e({\x8e\'|\x8e\'}\x8e\'~\x8e&\x7f\x8e&\x80\x8e&\x81\x8e%\x82\x8e%\x83\x8d$\x84\x8d$\x85\x8d$\x86\x8d#\x87\x8d#\x88\x8d#\x89\x8d"\x89\x8d"\x8a\x8d"\x8b\x8d!\x8c\x8d!\x8d\x8c!\x8e\x8c \x8f\x8c \x90\x8c \x91\x8c\x1f\x92\x8c\x1f\x93\x8b\x1f\x94\x8b\x1f\x95\x8b\x1f\x96\x8b\x1e\x97\x8a\x1e\x98\x8a\x1e\x99\x8a\x1e\x99\x8a\x1e\x9a\x89\x1e\x9b\x89\x1e\x9c\x89\x1e\x9d\x88\x1e\x9e\x88\x1e\x9f\x88\x1e\xa0\x87\x1f\xa1\x87\x1f\xa2\x86\x1f\xa3\x86 \xa4\x85 \xa5\x85!\xa6\x85!\xa7\x84"\xa7\x84#\xa8\x83#\xa9\x82$\xaa\x82%\xab\x81&\xac\x81\'\xad\x80(\xae\x7f)\xaf\x7f*\xb0~+\xb1},\xb1}.\xb2|/\xb3{0\xb4z2\xb5z3\xb6y5\xb7x6\xb8w8\xb9v9\xb9v;\xbau=\xbbt>\xbcs@\xbdrB\xbeqD\xbepE\xbfoG\xc0nI\xc1mK\xc2lM\xc2kO\xc3iQ\xc4hS\xc5gU\xc6fW\xc6eY\xc7d[\xc8b^\xc9a`\xc9`b\xca_d\xcb]g\xcc\\i\xcc[k\xcdYm\xceXp\xceVr\xcfUt\xd0Tw\xd0Ry\xd1Q|\xd2O~\xd2N\x81\xd3L\x83\xd3K\x86\xd4I\x88\xd5G\x8b\xd5F\x8d\xd6D\x90\xd6C\x92\xd7A\x95\xd7?\x97\xd8>\x9a\xd8<\x9d\xd9:\x9f\xd98\xa2\xda7\xa5\xda5\xa7\xdb3\xaa\xdb2\xad\xdc0\xaf\xdc.\xb2\xdd,\xb5\xdd+\xb7\xdd)\xba\xde\'\xbd\xde&\xbf\xdf$\xc2\xdf"\xc5\xdf!\xc7\xe0\x1f\xca\xe0\x1e\xcd\xe0\x1d\xcf\xe1\x1c\xd2\xe1\x1b\xd4\xe1\x1a\xd7\xe2\x19\xda\xe2\x18\xdc\xe2\x18\xdf\xe3\x18\xe1\xe3\x18\xe4\xe3\x18\xe7\xe4\x19\xe9\xe4\x19\xec\xe4\x1a\xee\xe5\x1b\xf1\xe5\x1c\xf3\xe5\x1e\xf6\xe6\x1f\xf8\xe6!\xfa\xe6"\xfd\xe7$'


def get_heatmap_color(val):
    """Convert 0-1 value into #RRGGBB color."""
    val = int(val * 255)
    v = [VIRIDIS_COLORS[3 * val + i] for i in range(3)]
    return f"#{v[0]:02X}{v[1]:02X}{v[2]:02X}"



from PySide2.QtCore import (
    Property,
    QAbstractListModel,
    QModelIndex,
    QObject,
    Qt,
    Signal,
    Slot,
)

from virtual_warehouse.environment import LOCATION_TYPE_MAP
from virtual_warehouse.heatmap import get_heatmap_color


class SingleLocation(QObject):
    """Class representing single location in warehouse."""

    def __init__(self, location):
        self._i = location
        self._color = LOCATION_TYPE_MAP[self._i.has_ltype]["color"]
        self._gcolor = LOCATION_TYPE_MAP[self._i.has_ltype]["gray_color"]
        self._mesh_file = LOCATION_TYPE_MAP[self._i.has_ltype]["mesh"]
        # Required for displaying selected locations
        self.names = [self._i.name]

    def get_dict(self):
        """Return dictionary representing the location."""
        return {
            "name": self._i.name,
            "type": self._i.has_ltype,
            "mesh_file": self._mesh_file,
            "color": self._color,
            "gray_color": self._gcolor,
            "x": self._i.has_x,
            "y": self._i.has_y,
            "z": self._i.has_z,
            "length": self._i.has_length,
            "width": self._i.has_width,
            "height": self._i.has_height,
        }

    def get_heat(self):
        return self._i.has_freq


class MultiLocation(QObject):
    """Class representing multiple locations merged together."""

    def __init__(self, locations):
        self._l = locations
        self._i = locations[0]
        self._color = LOCATION_TYPE_MAP[self._i.has_ltype]["color"]
        self._gcolor = LOCATION_TYPE_MAP[self._i.has_ltype]["gray_color"]
        self._mesh_file = LOCATION_TYPE_MAP[self._i.has_ltype]["mesh"]
        # Required for displaying selected locations
        self.names = [i.name for i in locations]

    def get_dict(self):
        """Return dictionary representing the multi-location."""
        return {
            "name": self._i.name,
            "type": self._i.has_ltype,
            "mesh_file": self._mesh_file,
            "color": self._color,
            "gray_color": self._gcolor,
            "x": self._i.has_x,
            "y": self._i.has_y,
            "z": self._i.has_z,
            "length": self._i.has_length,
            "width": self._i.has_width,
            "height": self._i.has_height,
        }

    def get_heat(self, level=-1):
        return (
            self._l[level].has_freq if level != -1 else sum(i.has_freq for i in self._l)
        )


class UniversalLocationListModel(QAbstractListModel):
    """List model providing locations for 2D or 3D view."""

    ObjectRole = Qt.UserRole + 1

    maxChanged = Signal()
    levelChanged = Signal()

    def __init__(self, on_change, objects=None, parent=None):
        super(UniversalLocationListModel, self).__init__(parent)
        self._on_change = on_change
        self.set_data({} if objects is None else objects)
        self._max_heat = 0
        self._level = -1
        self._max_level = 1

    def set_data(self, objects):
        self._objects = objects
        self._keys = list(objects.keys())
        self.name_to_idx = {}
        for i, k in enumerate(self._keys):
            for n in self._objects[k].names:
                self.name_to_idx[n] = i

        if objects:
            self.update_max_heat()
            self._max_level = max(l._i.has_z for l in self._objects.values())
            self.maxChanged.emit()

    def update_max_heat(self):
        if len(self._objects) == 0:
            return 0
        self._max_heat = max(o.get_heat() for _, o in self._objects.items())

    def get_idx(self, idx):
        return self._objects[self._keys[idx]]

    @Property(float, constant=False, notify=maxChanged)
    def max_heat(self):
        return self._max_heat

    @Property(int, constant=False, notify=maxChanged)
    def max_level(self):
        return self._max_level

    @Slot(result=int)
    def rowCount(self, parent=QModelIndex()):
        return len(self._keys)

    @Slot(int, result="QVariant")
    def get(self, index):
        return self.get_idx(index).get_dict()

    @Property(int, constant=False, notify=levelChanged)
    def level(self):
        return self._level

    @level.setter
    def set_level(self, level):
        self._level = level

    @Slot(int, float, result=str)
    def get_heat(self, index, max_heat=None):
        freq = self.get_idx(index).get_heat(self._level)
        if max_heat:
            return get_heatmap_color(freq / max_heat)
        return get_heatmap_color(freq / self._max_heat)

    def data(self, index, role):
        if index.isValid() and role == UniversalLocationListModel.ObjectRole:
            return self.get_idx(index.row())
        return None

    def roleNames(self):  # skipcq: PYL-R0201
        """Set role names, QAbstractListModel requires implementation of this method."""
        roles = {}
        roles[UniversalLocationListModel.ObjectRole] = b"object"
        return roles



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['virtual_warehouse', 'virtual_warehouse.parser']

package_data = \
{'': ['*'],
 'virtual_warehouse': ['resources/images/*',
                       'resources/images/textures/*',
                       'resources/objects/*',
                       'resources/qml/*']}

install_requires = \
['Owlready2==0.29',
 'PySide2==5.15.0',
 'openpyxl>=3.0.7,<4.0.0',
 'rdflib>=5.0.0,<6.0.0',
 'xlrd==1.2.0']

entry_points = \
{'console_scripts': ['virtual-warehouse = virtual_warehouse.__main__:main']}

setup_kwargs = {
    'name': 'virtual-warehouse',
    'version': '0.0.1',
    'description': 'Warehouse visualisation app',
    'long_description': '# Virtual Warehouse\n[![Documentation Status](https://readthedocs.org/projects/virtual-warehouse/badge/?version=latest)](https://virtual-warehouse.readthedocs.io/en/latest/?badge=latest) \n[![DeepSource](https://deepsource.io/gh/Breta01/virtual-warehouse.svg/?label=active+issues&show_trend=true)](https://deepsource.io/gh/Breta01/virtual-warehouse/?ref=repository-badge)\n\nVisualisation of warehouse\n',
    'author': 'Břetislav Hájek',
    'author_email': 'info@bretahajek.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Breta01/virtual-warehouse',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<3.9',
}


setup(**setup_kwargs)



"""Different environment settings."""

# TODO: Add correct objects
LOCATION_TYPE_MAP = {
    "floor": {"color": "gray", "gray_color": "gray", "mesh": ":/objects/floor.obj",},
    "rack": {"color": "#62ca5f", "gray_color": "gray", "mesh": ":/objects/rack.obj",},
    "wall": {"color": "black", "gray_color": "black", "mesh": ":/objects/floor.obj",},
    "inbound_door": {
        "color": "#62ca5f",
        "gray_color": "#222",
        "mesh": ":/objects/floor.obj",
    },
    "outbound_door": {
        "color": "#fd8a24",
        "gray_color": "#aaa",
        "mesh": ":/objects/floor.obj",
    },
    "staging_area": {
        "color": "#2c728e",
        "gray_color": "#ddd",
        "mesh": ":/objects/floor.obj",
    },
    "custom": {
        "color": "#fde724",
        "gray_color": "white",
        "mesh": ":/objects/floor.obj",
    },
}



"""Main app module starting the Virtual Warehouse app."""
import sys

from PySide2 import QtCore
from PySide2.QtGui import QGuiApplication, QIcon
from PySide2.QtQml import QQmlApplicationEngine

import virtual_warehouse.main_rc  # skipcq: PYL-W0611
from virtual_warehouse import __version__
from virtual_warehouse.view_controller import ViewController

try:
    # Win incon support, include in try/except block if you're also targeting Mac/Linux
    from PySide2.QtWinExtras import QtWin

    myappid = "com.bretahajek.virtual-varehouse"
    QtWin.setCurrentProcessExplicitAppUserModelID(myappid)
except ImportError:
    pass


def run():
    """Start Virtual Warehouse application."""
    app = QGuiApplication(sys.argv)
    app.setOrganizationName("Bretislav Hajek")
    app.setOrganizationDomain("bretahajek.com")
    app.setApplicationName("Virtual Warehouse")
    app.setApplicationVersion("0.1")
    engine = QQmlApplicationEngine()

    controller = ViewController()
    engine.rootContext().setContextProperty("ViewController", controller)
    engine.rootContext().setContextProperty("versionNumber", __version__)

    app.setWindowIcon(QIcon(":/images/icon.png"))
    engine.load(QtCore.QUrl("qrc:/qml/main.qml"))

    if not engine.rootObjects():
        sys.exit(-1)
    sys.exit(app.exec_())



"""Data model using Owlready2 ontology."""
import datetime
from typing import List

from owlready2 import (
    DataProperty,
    FunctionalProperty,
    ObjectProperty,
    Thing,
    default_world,
    destroy_entity,
    get_ontology,
)

from virtual_warehouse.parser.utils import (
    convert_date,
    convert_dim,
    convert_type,
    convert_weight,
)

# We could use actual irl with already created ontology
BASE_IRI = "http://warehouse/onto.owl"
onto = get_ontology(BASE_IRI)


def destroy_all(cls):
    """Destroy all instances of given class.

    Args:
        cls (class): ontology class
    """
    for i in cls.instances():
        destroy_entity(i)


with onto:

    class Country(Thing):
        """Class representing different countries.

        Attributes:
            name (str): Name of the country
        """

        pass

    class Inventory(Thing):
        """Description of inventory balance for given date and location.
        TODO: Re-work the structure of inventory saving (individual items)

        Attributes:
            name (str): unique ID of inventory object
            has_date (datetime.datetime): date of the inventory status
            has_location (Location): location object related to this inventory status
            has_ltype (str): location type
            has_items (List[Item]): list of items on given location
            has_expiry_date (datetime.datetime): date of the item expiration
            has_availabel_qty (int): availabel quantity of the item
            has_onhand_qty (int): on-hand quantity of the item
            has_transi_qty (int): transiting quantity of the item
            has_allocated_qty (int): allocated quantity of the item
            has_suspense_qty (int): suspended quantity of the item
        """

        @classmethod
        def create(
            cls,
            date: str,
            location_id: str,
            ltype: str,
            item_id: str,
            expiry_date: str,
            available_qty: int,
            onhand_qty: int,
            transi_qty: int,
            allocated_qty: int,
            suspense_qty: int,
        ):
            date_t = convert_date(date, "%d.%m.%Y")
            expiry_date = convert_date(expiry_date, "%d.%m.%Y")

            item = onto.search_one(iri=f"{BASE_IRI}#{item_id}")
            location = onto.search_one(iri=f"{BASE_IRI}#{location_id}")

            return cls(
                f"{date}-{location_id}",
                has_date=date_t,
                has_location=location,
                has_ltype=ltype,
                has_items=[item],
                has_expiry_date=expiry_date,
                has_availabel_qty=available_qty,
                has_onhand_qty=onhand_qty,
                has_transi_qty=transi_qty,
                has_allocated_qty=allocated_qty,
                suspense_qty=suspense_qty,
            )

        def add_item(self, item_id: str):
            item = onto.search_one(iri=f"{BASE_IRI}#{item_id}")
            self.has_items.append(item)

        @classmethod
        def destroy_all(cls):
            """Destroy all instances of Inventory class."""
            destroy_all(cls)

    class ItemUnit(Thing):
        """Represents different packaging units of item.

        Attributes:
            name (str): unique id (item_id-unit_lvl)
            has_conversion_qty (int): number of packages inside base unit
            has_qty_uom (str): name of unit
            has_length (float): length of packaging (in meters)
            has_width (float): width of packaging (in meters)
            has_height (float): height of packaging (in meters)
            has_weight (float): weight of packaging (in kg)
        """

        @classmethod
        def create(
            cls,
            _id: str,
            conversion_qty: int,
            qty_uom: str,
            length: float,
            width: float,
            height: float,
            dim_uom: str,
            weight: float,
            weight_uom: str,
        ):
            length = convert_dim(length, dim_uom)
            width = convert_dim(width, dim_uom)
            height = convert_dim(height, dim_uom)
            weight = convert_weight(weight, weight_uom)

            return cls(
                _id,
                has_conversion_qty=conversion_qty,
                has_qty_uom=qty_uom,
                has_length=length,
                has_width=width,
                has_height=height,
                has_weight=weight,
            )

    class Item(Thing):
        """Basic description of item, including packaging units.

        Attributes:
            name (str): id of item
            has_description (str): description of items
            has_gtype (str): type of goods
            has_zone (str): required zone for storing item
            has_base_unit (ItemUnit): base packaging unit
            has_unit_levels (List[ItemUnit]): list of all different packaging units
        """

        @classmethod
        def create(
            cls,
            _id: str,
            description: str,
            gtype: str,
            zone: str,
            base_unit: ItemUnit,
            unit_levels: List[ItemUnit],
        ):
            return cls(
                _id,
                has_description=description,
                has_gtype=gtype,
                has_zone=zone,
                has_base_unit=base_unit,
                has_unit_levels=unit_levels,
            )

        @classmethod
        def destroy_all(cls):
            """Destroy all instances of Item class as well as related entities."""
            destroy_all(cls)
            destroy_all(ItemUnit)

        @staticmethod
        def get_by_locations(locations):
            """Get list of items stored at given location.

            Args:
               locations (List[Location]): list of locations to look for.

            Returns:
               List[Item]: list of items stored at locations
            """
            # TODO: Check inventory date
            values = " ".join(map(lambda x: f"<{x.iri}>", locations))
            i1 = "VALUES ?loc { " + values + " } .\n"
            i2 = f"?inv <{has_location.iri}> ?loc .\n"
            i3 = f"?inv <{has_items.iri}> ?item .\n"

            query = "SELECT DISTINCT ?item WHERE { " + i1 + i2 + i3 + " }"
            return list(default_world.sparql_query(query))

        @staticmethod
        def get_by_orders(orders):
            """Get list of items included in given orders.

            Args:
               orders (List[Order]): list of orders to search.

            Returns:
               List[Item]: list of items included in orders.
            """
            values = " ".join(map(lambda x: f"<{x.iri}>", orders))
            i1 = "VALUES ?ord { " + values + " } .\n"
            i2 = f"?ord <{has_ordered_items.iri}> ?oi .\n"
            i3 = f"?oi <{has_item.iri}> ?item .\n"

            query = "SELECT DISTINCT ?item WHERE { " + i1 + i2 + i3 + " }"
            return list(default_world.sparql_query(query))

    class Location(Thing):
        """Description of location in warehouse.

        Attributes:
            name (str): unique id of location
            has_ltype (str): type of location
            has_lclass (str): class of location
            has_lsubclass (str): subclass of location
            has_length (float): length of location
            has_width (float): width of location
            has_height (float): height of location
            has_max_weight (float): maximal weight which can location hold
            has_zone (str): zone where is location located
            has_x (float): x coordinate of location
            has_y (float): y coordinate of location
            has_z (float): z coordinate of location
            has_freq (int): frequency calculated for heat map displaying
        """

        @classmethod
        def create(
            cls,
            _id: str,
            ltype: str,
            lclass: str,
            lsubclass: str,
            length: float,
            width: float,
            height: float,
            dim_uom: str,
            max_weight: float = None,
            weight_uom: str = None,
            zone: str = None,
            # Params which are usually calculated later:
            x: float = None,
            y: float = None,
            z: float = None,
            freq: int = 0,
        ):
            ltype = convert_type(ltype)
            length = convert_dim(length, dim_uom)
            width = convert_dim(width, dim_uom)
            height = convert_dim(height, dim_uom)
            if max_weight:
                max_weight = convert_weight(max_weight, weight_uom)

            return cls(
                _id,
                has_ltype=ltype,
                has_lclass=lclass,
                has_lsubclass=lsubclass,
                has_length=length,
                has_width=width,
                has_height=height,
                has_max_weight=max_weight,
                has_zone=zone,
                has_x=x,
                has_y=y,
                has_z=z,
                has_freq=freq,
            )

        def set_coord(self, x: float, y: float, z: float):
            """Additionally set coordinates of the location."""
            self.has_x, self.has_y, self.has_z = x, y, z  # skipcq: PYL-W0201

        def get_2d(self):
            """Get planar coordinates of location (used for top-down view)."""
            return (self.has_x, self.has_y)

        @classmethod
        def destroy_all(cls):
            """Destroy all instances of Location class."""
            destroy_all(cls)

        @staticmethod
        def get_by_orders(orders):
            """Get list of locations which are potentially accessed by given orders.

            Args:
               locations (List[Order]): list of locations to inspect

            Returns:
               List[Location]: list of locations containing ordered items
            """
            values = " ".join(map(lambda x: f"<{x.iri}>", orders))
            i1 = "VALUES ?ord { " + values + " } .\n"
            i2 = f"?ord <{has_ordered_items.iri}> ?oi .\n"
            i3 = f"?oi <{has_item.iri}> ?item .\n"
            i4 = f"?inv <{has_items.iri}> ?item .\n"
            i5 = f"?inv <{has_location.iri}> ?loc .\n"

            query = "SELECT DISTINCT ?loc WHERE { " + i1 + i2 + i3 + i4 + i5 + " }"
            return list(default_world.sparql_query(query))

        @staticmethod
        def get_by_items(items):
            """Get list of locations storing given items.

            Args:
               items (List[Item]): list of items to locate items.

            Returns:
               List[Location]: list of locations storing items
            """
            values = " ".join(map(lambda x: f"<{x.iri}>", items))
            i1 = "VALUES ?item { " + values + " } .\n"
            i2 = f"?inv <{has_items.iri}> ?item .\n"
            i3 = f"?inv <{has_location.iri}> ?loc .\n"

            query = "SELECT DISTINCT ?loc WHERE { " + i1 + i2 + i3 + " }"
            return list(default_world.sparql_query(query))

    class OrderedItem(Thing):
        """Description of item instance in order.

        Attributes:
            name (str): id of object (ordered_id-item_id)
            has_item (Item): item object instantiated by this order
            has_requested_qty (int): requested amount of item inside ordered
            has_total_qty (int): amount of item provided inside order
            qty_uom (int): unit of measure of quantities
        """

        pass

    class Order(Thing):
        """Description of single order from warehouse.

        Attributes:
            name (str): id of the order
            has_direction (str): outbound/inbound
            has_country (str): Country object representing shipping destination
            has_delivery_date (datetime.datetime): requested delivery date
            has_s_ship_date (datetime.datetime): scheduled sipping date
            has_a_ship_date (datetime.datetime): actual sipping date
            has_line_num (int): ERP order line number
            has_ordered_items (List[OrderedItem]): list of item instances inside order
        """

        @classmethod
        def create(
            cls,
            _id: str,
            direction: str,
            country_id: str,
            delivery_date: str,
            s_ship_date: str,
            a_ship_date: str,
            line_num: int,
            item_id: str,
            requested_qty: int,
            total_qty: int,
            qty_uom: str,
        ):
            _id = str(_id)
            # Convert to datetime
            delivery_date = convert_date(delivery_date, "%d.%m.%Y")
            s_ship_date = convert_date(s_ship_date, "%d.%m.%Y")
            a_ship_date = convert_date(a_ship_date, "%d.%m.%Y")

            country = onto.search_one(iri=f"{BASE_IRI}#{country_id}")
            if country is None:
                country = Country(country_id)

            item = onto.search_one(iri=f"{BASE_IRI}#{item_id}")
            oi = OrderedItem(
                f"{_id}-{item_id}",
                has_item=item,
                has_requested_qty=requested_qty,
                has_total_qty=total_qty,
                has_qty_uom=qty_uom,
            )

            return cls(
                _id,
                has_direction=direction,
                has_country=country,
                has_delivery_date=delivery_date,
                has_s_ship_date=s_ship_date,
                has_a_ship_date=a_ship_date,
                has_line_num=line_num,
                has_ordered_items=[oi],
            )

        def add_item(
            self, item_id: str, requested_qty: int, total_qty: int, qty_uom: str
        ):
            """Create and add item instance into order."""
            item = onto.search_one(iri=f"{BASE_IRI}#{item_id}")
            oi = OrderedItem(
                f"{self.name}-{item_id}",
                has_item=item,
                has_requested_qty=requested_qty,
                has_total_qty=total_qty,
                has_qty_uom=qty_uom,
            )
            self.has_ordered_items.append(oi)

        @classmethod
        def destroy_all(cls):
            """Destroy all instances of Order class as well as related entities."""
            destroy_all(cls)
            destroy_all(OrderedItem)

        @staticmethod
        def get_by_items(items):
            """Get list of orders which contains given items.

            Args:
               items (List[Item]): list of items to look for.

            Returns:
               List[Order]: list of orders containing at leas one of the provided items
            """
            values = " ".join(map(lambda x: f"<{x.iri}>", items))
            i1 = "VALUES ?item { " + values + " } .\n"
            i2 = f"?oi <{has_item.iri}> ?item .\n"
            i3 = f"?ord <{has_ordered_items.iri}> ?oi .\n"

            query = "SELECT DISTINCT ?ord WHERE { " + i1 + i2 + i3 + " }"
            return list(default_world.sparql_query(query))

        @staticmethod
        def get_by_locations(locations):
            """Get list of orders which potentially access given locations.

            Args:
               locations (List[Location]): list of locations to inspect

            Returns:
               List[Order]: list of orders containing items stored at given locations
            """
            values = " ".join(map(lambda x: f"<{x.iri}>", locations))
            i1 = "VALUES ?loc { " + values + " } .\n"
            i2 = f"?inv <{has_location.iri}> ?loc .\n"
            i3 = f"?inv <{has_items.iri}> ?item .\n"
            i4 = f"?oi <{has_item.iri}> ?item .\n"
            i5 = f"?ord <{has_ordered_items.iri}> ?oi .\n"

            query = "SELECT DISTINCT ?ord WHERE { " + i1 + i2 + i3 + i4 + i5 + " }"
            return list(default_world.sparql_query(query))

    # Order properties
    class has_direction(DataProperty, FunctionalProperty):
        domain = [Order]
        range = [str]

    class has_country(ObjectProperty, FunctionalProperty):
        domain = [Order]
        range = [Country]

    class has_delivery_date(DataProperty, FunctionalProperty):
        domain = [Order]
        range = [datetime.datetime]

    class has_s_ship_date(DataProperty, FunctionalProperty):
        domain = [Order]
        range = [datetime.date]

    class has_a_ship_date(DataProperty, FunctionalProperty):
        domain = [Order]
        range = [datetime.date]

    class has_line_num(DataProperty, FunctionalProperty):
        domain = [Order]
        range = [int]

    class has_ordered_items(Order >> OrderedItem):
        pass

    # OrderedItem properties
    class has_item(ObjectProperty, FunctionalProperty):
        domain = [OrderedItem]
        range = [Item]

    class in_order(OrderedItem >> Order):
        inverse_property = has_ordered_items

    class has_requested_qty(DataProperty, FunctionalProperty):
        domain = [OrderedItem]
        range = [int]

    class has_total_qty(DataProperty, FunctionalProperty):
        domain = [OrderedItem]
        range = [int]

    class has_qty_uom(DataProperty, FunctionalProperty):
        domain = [OrderedItem, ItemUnit]
        range = [str]

    # ItemUnit properties
    class has_conversion_qty(DataProperty, FunctionalProperty):
        domain = [ItemUnit]
        range = [int]

    class has_weight(DataProperty, FunctionalProperty):
        domain = [ItemUnit]
        range = [float]

    # Item properties
    class in_ordered_item(Item >> OrderedItem):
        inverse_property = has_item

    class has_description(DataProperty, FunctionalProperty):
        domain = [Item]
        range = [str]

    class has_gtype(DataProperty, FunctionalProperty):
        domain = [Item]
        range = [str]

    class has_zone(DataProperty, FunctionalProperty):
        domain = [Item, Location]
        range = [str]

    class has_base_unit(ObjectProperty, FunctionalProperty):
        domain = [Item]
        range = [ItemUnit]

    class has_unit_levels(Item >> ItemUnit):
        pass

    # Location properties
    class has_ltype(DataProperty, FunctionalProperty):
        domain = [Location, Inventory]
        range = [str]

    class has_lclass(DataProperty, FunctionalProperty):
        domain = [Location]
        range = [str]

    class has_lsubclass(DataProperty, FunctionalProperty):
        domain = [Location]
        range = [str]

    class has_length(DataProperty, FunctionalProperty):
        domain = [Location, ItemUnit]
        range = [float]

    class has_width(DataProperty, FunctionalProperty):
        domain = [Location, ItemUnit]
        range = [float]

    class has_height(DataProperty, FunctionalProperty):
        domain = [Location, ItemUnit]
        range = [float]

    class has_max_weight(DataProperty, FunctionalProperty):
        domain = [Location]
        range = [float]

    class has_coord(DataProperty, FunctionalProperty):
        domain = [Location]
        range = [str]

    class has_freq(DataProperty, FunctionalProperty):
        domain = [Location]
        range = [int]

    class has_x(DataProperty, FunctionalProperty):
        domain = [Location]
        range = [float]

    class has_y(DataProperty, FunctionalProperty):
        domain = [Location]
        range = [float]

    class has_z(DataProperty, FunctionalProperty):
        domain = [Location]
        range = [float]

    # Inventory properties
    class has_date(DataProperty, FunctionalProperty):
        domain = [Inventory]
        range = [datetime.datetime]

    class has_location(ObjectProperty, FunctionalProperty):
        domain = [Inventory]
        range = [Location]

    class has_items(Inventory >> Item):
        pass

    class has_expiry_date(DataProperty, FunctionalProperty):
        domain = [Inventory]
        range = [datetime.datetime]

    class has_availabel_qty(DataProperty, FunctionalProperty):
        domain = [Inventory]
        range = [int]

    class has_onhand_qty(DataProperty, FunctionalProperty):
        domain = [Inventory]
        range = [int]

    class has_transi_qty(DataProperty, FunctionalProperty):
        domain = [Inventory]
        range = [int]

    class has_allocated_qty(DataProperty, FunctionalProperty):
        domain = [Inventory]
        range = [int]

    class has_supsense_qty(DataProperty, FunctionalProperty):
        domain = [Inventory]
        range = [int]


# sync_reasoner(infer_property_values=True)

# SQLite3
# default_world.graph.db
# default_world.graph.execute
